from typing import Tuple, List

from .extract import extract

__all__ = ["replace"]


def replace(text: str, pattern: str) -> Tuple[str, List[str]]:
    """
    Replace all sigils in the text with another pattern.
    Returns the replaced text and a list of sigils in found order.
    This will not resolve the sigils.

    :param text: The text with the sigils to be replaced.
    :param pattern: A string used to replace the sigils with.
    :return: A tuple of: replaced text, list of sigil strings.

    >>> # Protect against SQL injection.
    >>> replace("select * from users where username = [USER]", "?")
    ('select * from users where username = ?', ['[USER]'])
    """

    sigils = list(extract(text))
    for sigil in set(sigils):
        text = text.replace(sigil, pattern)
    return text, sigils



import logging
from typing import Any

from .extract import extract
from .parse import parser
from .transform import SigilResolver
from .exceptions import SigilError

logger = logging.getLogger(__name__)

__all__ = ["set_context", "resolve"]

# Global default context
# Don't modify this directly, use set_context()
_context = {}


# noinspection PyUnresolvedReferences
def set_context(key: str, value: Any) -> None:
    """
    Set a global default context for resolve.

    :param key: String key used to lookup context.
    :param value: The context value, usually a callable or instance.

    >>> # Use resolve with only default global context
    >>> set_context("SYS", {"PATH": "/usr/bin"})
    >>> resolve("System path is [SYS.PATH]")
    'System path is /usr/bin'

    >>> # Use set_context to create a global function
    >>> set_context("PERCENT", lambda x: int(float(x) * 100))
    >>> context = {"VALUE": 0.5}
    >>> resolve("The percent is [VALUE.PERCENT]%", context)
    'The percent is 50%'
    """

    global _context
    _context[key] = value


# noinspection PyBroadException
def resolve(text: str, context: dict = None, required=False) -> str:
    """
    Resolve all sigils found in text, using the specified context.
    If the sigil can't be resolved, return it unchanged unless
    required is True, in which case raise SigilError.

    :param text: The text containing sigils.
    :param context: Optional dict of context used for resolution.
    :param required: If True, raise SigilError if a sigil can't resolve.

    >>> # Resolving sigils using local context:
    >>> context = {"ENV": {"HOST": "localhost"}, "USER": "arthexis"}
    >>> resolve("Connect to [ENV.HOST] as [USER]", context)
    'Connect to localhost as arthexis'
    """
    global _context

    context = {**_context, **(context or {})}
    if not context:
        raise ValueError("context is required")
    if text.startswith('[') and text.endswith(']'):
        sigils = {text}
    else:
        sigils = set(extract(text))
    if not sigils:
        logger.debug("No sigils found in '%s'", text)
        return text
    logger.debug("Found sigils: %s", sigils)
    for sigil in sigils:
        try:
            tree = parser.parse(sigil)
            value = SigilResolver(context).transform(tree).children[0]
            if callable(value):
                raise ValueError("Required arg or node missing.")
            logger.debug("Sigil %s resolved to '%s'", sigil, value)
            text = text.replace(sigil, repr(value))
        except Exception as ex:
            if required:
                raise SigilError(sigil) from ex
            logger.debug("Sigil %s not resolved", sigil)
    return text



import logging

from lark import Transformer, v_args

__all__ = ["SigilResolver"]

logger = logging.getLogger(__name__)


class SigilResolver(Transformer):

    def __init__(self, context: dict = None):
        super().__init__()
        self.ctx = context or {}

    @v_args(inline=True)
    def sigil(self, *nodes):
        obj = None
        for node in nodes:
            key, arg = node
            if not obj:
                # We don't have a parent object: its the first node
                # Try finding the object in context only
                try:
                    obj = self.ctx[key]
                except KeyError as ex:
                    logger.error("Key %s not in context", key)
                    raise ex
            else:
                # We have a parent object: not the first node
                # Try the following in order:
                # 1. Find the object as an item in the parent
                # 2. Find the object as an attribute of the parent
                # 3. Find a function with that name in the context and
                #    call it by passing the parent as the argument
                # Make the result the new parent object
                try:
                    obj = obj[key]
                except KeyError as ex:
                    logger.error("Key %s not an item in %s", key, obj)
                    raise ex
                except TypeError as ex:
                    _key = key.casefold()
                    if key.startswith("_"):
                        raise AttributeError("Key %s cannot be private", key)
                    try:
                        obj = getattr(obj, _key)
                        continue
                    except AttributeError as ex:
                        func = self.ctx.get(key, None)
                        if callable(func):
                            logger.debug("Apply %s to %s", func, obj)
                            obj = func(obj)
                            continue
                        logger.error("Key %s not an attribute of %s", key, obj)
                        raise ex
            if arg:
                # If and only if an arg was provided for this node
                # try calling the object with the provided arg, and make
                # the returned value the new parent object.
                try:
                    obj = obj(arg)
                except TypeError as ex:
                    logger.error("Key %s is not callable", key)
                    raise ex
        return obj

    # Flatten nodes (otherwise a Tree is returned)
    @v_args(inline=True)
    def node(self, key, arg=None):
        return [str(key), arg]

    # Flatten node args (otherwise a Tree is returned)
    @v_args(inline=True)
    def arg(self, value=None):
        return value



from typing import Iterator

__all__ = ["extract"]

_ld, _rd = "[", "]"


def extract(text: str) -> Iterator[str]:
    """
    Generator that extracts sigils from text.
    Works properly with nested sigils.

    :param text: The text to extract the sigils from.
    :return: An iterable of sigil strings.

    >>> # Extract simple sigils from text
    >>> list(extract("Connect to [ENV.HOST] as [USER]"))
    ['[ENV.HOST]', '[USER]']

    >>> # Nested sigils are allowed, only top level is extracted
    >>> list(extract("Environ [ENV=[SRC]]"))
    ['[ENV=[SRC]]']

    >>> # Malformed sigils are not extracted (no errors raised).
    >>> list(extract("Connect to [ENV.HOST as USER"))
    []
    """
    global _ld, _rd

    buffer, depth = [], 0
    for char in text:
        if char == _ld:
            depth += 1
        if depth > 0:
            buffer.append(char)
            if char == _rd:
                depth -= 1
                if depth == 0:
                    yield ''.join(buffer)
                    buffer.clear()



from lark import Lark

__all__ = ["parser"]

sigil_grammar = r"""
    start    : sigil
    sigil    : "[" node ("." node)* "]"
    node     : CNAME ["=" arg]
    arg      : sigil
             | "'" CNAME "'"
             | NUMBER

    %import common.CNAME
    %import common.NUMBER
    %import common.WS
    %ignore WS
"""

parser = Lark(sigil_grammar)



from setuptools import setup
from os import path

base_dir = path.abspath(path.dirname(__file__))
with open(path.join(base_dir, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='sigils',
    version='0.0.3',
    description='Extract, resolve and replace [SIGILS] embedded in text.',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url='http://github.com/arthexis/sigils',
    download_url='https://github.com/arthexis/sigils/archive/v0.0.3.tar.gz',
    author='Rafael Guillén',
    author_email='arthexis@gmail.com',
    license='MIT',
    keywords=["UTILS", "SIGIL", "STRING", "TEXT"],
    packages=['sigils'],
    zip_safe=True,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'Topic :: Text Processing',
        'Topic :: Utilities',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    install_requires=[
        'lark-parser',
    ],
    extras_require={
        'django': [
            'django',
        ],
        'dev': [
            'pytest',
            'black',
            'pytest-cov',
            'Sphinx',
        ]
    }
)
